from typing import Protocol

from turu.core.protocols.cursor import CursorProtocol


class ConnectionProtocol(Protocol):
    def close(self) -> None:
        """Close the connection now."""
        ...

    def commit(self) -> None:
        """Commit any pending transaction to the database."""
        ...

    def rollback(self) -> None:
        """Roll back to the start of any pending transaction.

        Closing a connection without committing the changes first
        will cause an implicit rollback to be performed.
        """
        ...

    def cursor(self) -> CursorProtocol:
        """Return a new Cursor Object using the connection."""
        ...



from contextlib import _AsyncGeneratorContextManager, _GeneratorContextManager
from pathlib import Path
from typing import (
    AsyncGenerator,
    Generator,
    Union,
    cast,
    overload,
)

import turu.core.async_cursor
import turu.core.cursor
from turu.core.record.async_record_cursor import AsyncRecordCursor as AsyncRecordCursor
from turu.core.record.csv_recorder import CsvRecorder as CsvRecorder
from turu.core.record.csv_recorder import CsvRecorderOptions as CsvRecorderOptions
from turu.core.record.record_cursor import RecordCursor as RecordCursor
from typing_extensions import Unpack, deprecated


@overload
def record_to_csv(
    record_filepath: Union[str, Path],
    cursor: turu.core.cursor.GenericCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> "_GeneratorContextManager[turu.core.cursor.GenericCursor]": ...


@overload
def record_to_csv(
    record_filepath: Union[str, Path],
    cursor: turu.core.async_cursor.GenericAsyncCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> _AsyncGeneratorContextManager[turu.core.async_cursor.GenericAsyncCursor]: ...


def record_to_csv(  # type: ignore
    record_filepath: Union[str, Path],
    cursor: Union[
        turu.core.cursor.GenericCursor, turu.core.async_cursor.GenericAsyncCursor
    ],
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
):
    """Records cursor's fetched data to CSV.

    Parameters:
        record_filepath: Path to record file.
        cursor: Cursor to record.
        enable: Enable recording.
        options: Options for CSV recorder.
    """
    if isinstance(cursor, turu.core.cursor.Cursor):
        return _GeneratorContextManager(
            _record_to_csv, (record_filepath, cursor), dict(enable=enable, **options)
        )

    elif isinstance(cursor, turu.core.async_cursor.AsyncCursor):
        return _AsyncGeneratorContextManager(
            _record_to_csv_async,
            (record_filepath, cursor),
            dict(enable=enable, **options),
        )

    raise NotImplementedError(f"cursor type {type(cursor)} is not supported")


def _record_to_csv(
    record_filepath: Union[str, Path],
    cursor: turu.core.cursor.GenericCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> Generator[turu.core.cursor.GenericCursor, None, None]:
    if isinstance(enable, str):
        enable = enable.lower() == "true"

    if enable:
        # NOTE: hack to get original cursor type hint.
        cursor = cast(
            turu.core.cursor.GenericCursor,
            getattr(cursor, "_RecordCursor", RecordCursor)(
                CsvRecorder(record_filepath, **options),
                cursor,
            ),
        )

    try:
        yield cursor

    finally:
        cursor.close()


async def _record_to_csv_async(
    record_filepath: Union[str, Path],
    cursor: turu.core.async_cursor.GenericAsyncCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> AsyncGenerator[turu.core.async_cursor.GenericAsyncCursor, None]:
    if isinstance(enable, str):
        enable = enable.lower() == "true"

    if enable:
        # NOTE: hack to get original cursor type hint.
        cursor = cast(
            turu.core.async_cursor.GenericAsyncCursor,
            getattr(cursor, "_AsyncRecordCursor", AsyncRecordCursor)(
                CsvRecorder(record_filepath, **options),
                cursor,
            ),
        )

    try:
        yield cursor

    finally:
        await cursor.close()


@overload
def record_as_csv(
    record_filepath: Union[str, Path],
    cursor: turu.core.cursor.GenericCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> "_GeneratorContextManager[turu.core.cursor.GenericCursor]": ...


@overload
def record_as_csv(
    record_filepath: Union[str, Path],
    cursor: turu.core.async_cursor.GenericAsyncCursor,
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
) -> _AsyncGeneratorContextManager[turu.core.async_cursor.GenericAsyncCursor]: ...


@deprecated(
    "This function is deprecated. Use `record_to_csv` instead.",
)
def record_as_csv(  # type: ignore
    record_filepath: Union[str, Path],
    cursor: Union[
        turu.core.cursor.GenericCursor, turu.core.async_cursor.GenericAsyncCursor
    ],
    *,
    enable: Union[str, bool, None] = True,
    **options: Unpack[CsvRecorderOptions],
):
    return record_to_csv(record_filepath, cursor, enable=enable, **options)



from typing import Any, List, Optional, Sequence, Tuple, Type, Union, overload

from turu.core.cursor import GenericRowType
from turu.core.mock.exception import (
    TuruMockResponseTypeMismatchError,
    TuruMockStoreDataNotFoundError,
)
from turu.core.tag import Tag


class TuruMockStore:
    def __init__(self):
        self._data: List[Tuple[Optional[Type], Union[Sequence, None, Exception]]] = []
        self._counter = 0

    def inject_operation_with_tag(
        self, tag: Type[Tag], exception: Optional[Exception] = None
    ):
        self._data.append((tag, exception))

    @overload
    def inject_response(
        self,
        row_type: None,
        response: Union[GenericRowType, Sequence[GenericRowType], None, Exception],
    ): ...

    @overload
    def inject_response(
        self,
        row_type: Type[GenericRowType],
        response: Union[GenericRowType, Sequence[GenericRowType], Exception],
    ): ...

    def inject_response(self, row_type, response):
        if row_type is not None and isinstance(response, row_type):
            response = (response,)

        self._data.append((row_type, response))

    def provide_response(
        self,
        row_type: Optional[Type],
    ) -> "Optional[Sequence[Any]]":
        self._counter += 1

        if len(self._data) == 0:
            raise TuruMockStoreDataNotFoundError(self._counter)

        store_row_type, store_response = self._data.pop(0)

        if store_row_type is not row_type and (str(row_type) != str(store_row_type)):
            raise TuruMockResponseTypeMismatchError(
                row_type, store_row_type, self._counter
            )

        if isinstance(store_response, Exception):
            raise store_response

        return store_response



from typing import Any, Generic, List, Optional, Protocol, Sequence, TypeVar

from typing_extensions import Self

Parameters = TypeVar("Parameters", contravariant=True)


class CursorProtocol(Generic[Parameters], Protocol):
    @property
    def rowcount(self) -> int:
        """The number of rows that the last `.execute*()` produced (for DQL statements like )
        or affected (for DML statements like or ).

        The attribute is `-1` in case no `.execute*()` has been performed
        on the cursor or the rowcount of the last operation is cannot be determined by the interface.
        """
        ...

    @property
    def arraysize(self) -> int:
        """The number of rows to fetch at a time with `.fetchmany()`.

        It defaults to 1 meaning to fetch a single row at a time.
        """
        ...

    @arraysize.setter
    def arraysize(self, size: int) -> None: ...

    def close(self) -> None: ...

    def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> Self:
        """Prepare and execute a database operation (query or command).

        Parameters:
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> Self:
        """Prepare a database operation (query or command)
        and then execute it against all parameter sequences or mappings.

        Parameters:
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    def fetchone(self) -> Optional[Any]:
        """Fetch the next row of a query result set."""
        ...

    def fetchmany(self, size: Optional[int] = None) -> List[Any]:
        """Fetch the next set of rows of a query result.

        An empty sequence is returned when no more rows are available.

        Parameters:
            size: The number of rows to fetch per call.
                    If this parameter is not used, it is usually refer to use the `.arraysize` attribute (Default is `1`).
                    If this parameter is used, then it is best for it to retain the same value from one `.fetchmany()` call to the next.
        """
        ...

    def fetchall(self) -> List[Any]:
        """Fetch all (remaining) rows of a query result"""
        ...

    def __iter__(self) -> Self: ...

    def __next__(self) -> Any: ...

    def __enter__(self) -> Self: ...

    def __exit__(self, exc_type, exc_value, traceback): ...



from abc import abstractmethod
from typing import (
    Generic,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
)

import turu.core.tag
from turu.core.cursor import GenericNewRowType as GenericNewRowType
from turu.core.cursor import GenericRowType as GenericRowType
from turu.core.protocols.async_cursor import AsyncCursorProtocol
from turu.core.protocols.async_cursor import Parameters as Parameters
from typing_extensions import Never, Self, override


class AsyncCursor(Generic[GenericRowType, Parameters], AsyncCursorProtocol[Parameters]):
    @property
    @abstractmethod
    def rowcount(self) -> int: ...

    @property
    @abstractmethod
    def arraysize(self) -> int: ...

    @arraysize.setter
    @abstractmethod
    def arraysize(self, size: int) -> None: ...

    @abstractmethod
    async def close(self) -> None: ...

    @abstractmethod
    async def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "AsyncCursor": ...

    @abstractmethod
    async def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> "AsyncCursor": ...

    @abstractmethod
    async def execute_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "AsyncCursor[GenericNewRowType, Parameters]":
        """
        Execute a database operation (query or command) and map each row to a `row_type`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    @abstractmethod
    async def executemany_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "AsyncCursor[GenericNewRowType, Parameters]":
        """Execute a database operation (query or command) against all parameter sequences or mappings.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    async def execute_with_tag(
        self,
        tag: Type[turu.core.tag.Tag],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "AsyncCursor[Never, Parameters]":
        """Execute a database operation (Insert, Update, Delete) with a tag.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),

        This method executes an operation (Insert, Update, Delete) that does not return a value with a tag.
        This tag is used to verify that the specified operation is executed in order when testing with Mock.
        """
        return await self.execute(operation, parameters)

    async def executemany_with_tag(
        self,
        tag: Type[turu.core.tag.Tag],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "AsyncCursor[Never, Parameters]":
        """Execute a database operation (Insert, Update, Delete) against all parameter sequences or mappings with a tag.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),

        This method executes an operation (Insert, Update, Delete) that does not return a value with a tag.
        This tag is used to verify that the specified operation is executed in order when testing with Mock.
        """
        return await self.executemany(operation, seq_of_parameters)

    @abstractmethod
    async def fetchone(self) -> Optional[GenericRowType]: ...

    @abstractmethod
    async def fetchmany(self, size: Optional[int] = None) -> List[GenericRowType]: ...

    @abstractmethod
    async def fetchall(self) -> List[GenericRowType]: ...

    @override
    def __aiter__(self) -> Self:
        return self

    @abstractmethod
    async def __anext__(self) -> GenericRowType: ...

    @override
    async def __aenter__(self):
        return self

    @override
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


GenericAsyncCursor = TypeVar("GenericAsyncCursor", bound=AsyncCursor)



from typing import Generic

from turu.core.cursor import GenericRowType


class Tag:
    """
    Base class for tags specified by `execute_with_tag`.
    """

    pass


class Insert(Tag, Generic[GenericRowType]):
    """
    Tag to specify when executing an INSERT query with `execute_with_tag`.

    It is assumed that a Pydantic model with table information is specified.

    ```python
    class User(pydantic.BaseModel):
        id: int
        name: str

    with database_conn.cursor() as cursor:
        cursor.execute_with_tag(
            tag.Insert[User],
            "INSERT INTO users (id, name) VALUES (1, 'Alice')"
        )
    ```
    """

    pass


class Update(Tag, Generic[GenericRowType]):
    """
    Tag to specify when executing an UPDATE query with `execute_with_tag`.

    It is assumed that a Pydantic model with table information is specified.

    ```python
    class User(pydantic.BaseModel):
        id: int
        name: str

    with database_conn.cursor() as cursor:
        cursor.execute_with_tag(
            tag.Update[User],
            "UPDATE users SET name = 'Bob' WHERE id = 1"
        )
    ```
    """

    pass


class Delete(Tag, Generic[GenericRowType]):
    """
    Tag to specify when executing a DELETE query with `execute_with_tag`.

    It is assumed that a Pydantic model with table information is specified.

    ```python
    class User(pydantic.BaseModel):
        id: int
        name: str

    with database_conn.cursor() as cursor:
        cursor.execute_with_tag(
            tag.Delete[User],
            "DELETE FROM users WHERE id = 1"
        )
    """

    pass


class Truncate(Tag, Generic[GenericRowType]):
    """
    Tag to specify when executing a TRUNCATE query with `execute_with_tag`.

    ```python
    with database_conn.cursor() as cursor:
        cursor.execute_with_tag(
            tag.Truncate,
            "TRUNCATE TABLE users"
        )
    """

    pass



import csv
from collections.abc import Callable
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, Sequence, TextIO, Union

import turu.core.cursor
from turu.core.exception import TuruRowTypeNotSupportedError
from turu.core.features import USE_PYDANTIC, PydanticModel
from turu.core.record.recorder_protcol import RecorderProtcol
from typing_extensions import LiteralString, NotRequired, TypedDict, Unpack


def not_supported(row: Any):
    raise TuruRowTypeNotSupportedError(type(row))


RecordCheckMap: Dict[str, Callable[[Any], bool]] = {
    "dataclass": is_dataclass,
    "tuple": lambda row: isinstance(row, tuple),
    "pydantic": lambda row: USE_PYDANTIC and isinstance(row, PydanticModel),
}

RecordHeaderMap: Dict[str, Callable[[Any], Iterable[str]]] = {
    "dataclass": lambda row: row.__dict__.keys(),
    "tuple": not_supported,
    "pydantic": lambda row: row.model_dump().keys(),
}

RecordRowMap: Dict[str, Callable[[Any], Iterable[str]]] = {
    "dataclass": lambda row: row.__dict__.values(),
    "tuple": lambda row: row,
    "pydantic": lambda row: row.model_dump().values(),
}

RecordRowsMap: Dict[str, Callable[[Any], Iterable[Iterable[str]]]] = {
    "dataclass": not_supported,
    "tuple": not_supported,
    "pydantic": not_supported,
}


class CsvRecorderOptions(TypedDict):
    """Options for CSV recorder."""

    header: NotRequired[bool]
    """Whether to write header or not."""

    limit: NotRequired[int]
    """Limit of rows to write."""


class CsvRecorder(RecorderProtcol):
    def __init__(
        self,
        filename: Union[str, Path],
        **options: Unpack[CsvRecorderOptions],
    ) -> None:
        self._file = Path(filename).open("w")
        self._writer = csv.writer(self._file)
        self._options = options
        self._writed_rowsize = 0

    @property
    def file(self) -> TextIO:
        return self._file

    def record(
        self,
        rows: Union[
            turu.core.cursor.GenericRowType,
            Sequence[turu.core.cursor.GenericRowType],
        ],
    ) -> None:
        try:
            datatype = get_datatype(rows)

        except TuruRowTypeNotSupportedError:
            pass

        for row in rows:  # type: ignore
            datatype = get_datatype(row)

            if self._writed_rowsize == 0:
                if self._options.get("header", True):
                    self._writer.writerow(RecordHeaderMap[datatype](row))

            if (limit := self._options.get("limit")) is not None:
                if self._writed_rowsize >= limit:
                    continue

            self._writed_rowsize += 1

            self._writer.writerow(RecordRowMap[datatype](row))

        return

    def close(self) -> None:
        if not self._file.closed:
            self._file.close()


def get_datatype(row: Any) -> str:
    for datatype, check in RecordCheckMap.items():
        if check(row):
            return datatype
    raise TuruRowTypeNotSupportedError(type(row))


def add_record_map(
    datatype: LiteralString,
    check_map: Callable[[Any], bool],
    header_map: Callable[[Any], Iterable[Any]],
    row_map: Callable[[Any], Iterable[Any]],
    rows_map: Callable[[Any], Iterable[Any]],
):
    global RecordHeaderMap, RecordRowMap

    RecordHeaderMap[datatype] = header_map
    RecordCheckMap[datatype] = check_map
    RecordRowMap[datatype] = row_map



from typing import Any, Iterator, List, Optional, Sequence, Tuple, Type

from turu.core.async_cursor import (
    AsyncCursor,
    GenericNewRowType,
    GenericRowType,
    Parameters,
)
from turu.core.mock.exception import (
    TuruMockUnexpectedFetchError,
)
from turu.core.mock.store import TuruMockStore
from turu.core.tag import Tag
from typing_extensions import Never, Self, override


class MockAsyncCursor(AsyncCursor[GenericRowType, Parameters]):
    def __init__(
        self,
        store: TuruMockStore,
        *,
        row_count: Optional[int] = None,
        rows_iter: Optional[Iterator] = None,
        row_type: Optional[Type[GenericRowType]] = None,
    ) -> None:
        self._turu_mock_store = store
        self._rowcount = row_count
        self._rows_iter = rows_iter
        self._row_type = row_type
        self._arraysize = 1

    @property
    def rowcount(self) -> int:
        return self._rowcount or -1

    @property
    def arraysize(self) -> int:
        return self._arraysize

    @arraysize.setter
    def arraysize(self, size: int) -> None:
        self._arraysize = size

    @override
    async def close(self) -> None:
        pass

    @override
    async def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "MockAsyncCursor[Tuple[Any], Parameters]":
        return self._make_new_mock_cursor(None)

    @override
    async def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> "MockAsyncCursor[Tuple[Any], Parameters]":
        return self._make_new_mock_cursor(None)

    @override
    async def execute_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "MockAsyncCursor[GenericNewRowType, Parameters]":
        return self._make_new_mock_cursor(row_type)

    @override
    async def executemany_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "MockAsyncCursor[GenericNewRowType, Parameters]":
        return self._make_new_mock_cursor(row_type)

    @override
    async def execute_with_tag(
        self, tag: type[Tag], operation: str, parameters: Optional[Parameters] = None
    ) -> "MockAsyncCursor[Never, Parameters]":
        return self._make_new_mock_cursor(None, tag)

    @override
    async def executemany_with_tag(
        self, tag: type[Tag], operation: str, seq_of_parameters: Sequence[Parameters]
    ) -> "MockAsyncCursor[Never, Parameters]":
        return self._make_new_mock_cursor(None, tag)

    @override
    async def fetchone(self) -> Optional[GenericRowType]:
        if self._rows_iter is None:
            return None

        try:
            return next(self._rows_iter)

        except (StopIteration, StopAsyncIteration):
            return None

    @override
    async def fetchmany(self, size: Optional[int] = None) -> List[GenericRowType]:
        if self._rows_iter is None:
            raise TuruMockUnexpectedFetchError()

        result = []
        for _ in range(size if size is not None else self.arraysize):
            try:
                result.append(self._rows_iter.__next__())
            except StopIteration:
                break

        return result

    @override
    async def fetchall(self) -> List[GenericRowType]:
        if self._rows_iter is None:
            raise TuruMockUnexpectedFetchError()

        return list(self._rows_iter)

    @override
    def __aiter__(self) -> Self:
        if self._rows_iter is None:
            raise TuruMockUnexpectedFetchError()
        return self

    @override
    async def __anext__(self) -> GenericRowType:
        if self._rows_iter is None:
            raise TuruMockUnexpectedFetchError()

        try:
            return next(self._rows_iter)
        except StopIteration as e:
            raise StopAsyncIteration from e

    def _make_new_mock_cursor(
        self,
        row_type: Optional[Type[GenericNewRowType]],
        tag: Optional[Type[Tag]] = None,
    ) -> "MockAsyncCursor":
        responses = self._turu_mock_store.provide_response(row_type or tag)

        if responses is None:
            return self.__class__(
                self._turu_mock_store,
                row_type=row_type,
            )

        else:
            return self.__class__(
                self._turu_mock_store,
                row_count=len(responses),
                rows_iter=iter(responses),
                row_type=row_type,
            )



from typing import (
    List,
    Optional,
    Sequence,
    Type,
    cast,
)

import turu.core.async_cursor
import turu.core.cursor
from turu.core.protocols.cursor import Parameters
from turu.core.record.recorder_protcol import RecorderProtcol


class AsyncRecordCursor(
    turu.core.async_cursor.AsyncCursor[turu.core.cursor.GenericRowType, Parameters]
):
    def __init__(
        self,
        recorder: RecorderProtcol,
        cursor: turu.core.async_cursor.AsyncCursor[
            turu.core.cursor.GenericRowType, Parameters
        ],
    ):
        self._recorder = recorder
        self.__record_taregt_cursor: turu.core.async_cursor.AsyncCursor = cursor

    @property
    def rowcount(self) -> int:
        return self.__record_taregt_cursor.rowcount

    @property
    def arraysize(self) -> int:
        return self.__record_taregt_cursor.arraysize

    @arraysize.setter
    def arraysize(self, size: int) -> None:
        self.__record_taregt_cursor.arraysize = size

    async def close(self) -> None:
        await self.__record_taregt_cursor.close()
        self._recorder.close()

    async def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "AsyncRecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        self.__record_taregt_cursor = await self.__record_taregt_cursor.execute(
            operation, parameters
        )

        return self

    async def executemany(
        self, operation: str, seq_of_parameters: "Sequence[Parameters]", /
    ) -> "AsyncRecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        self.__record_taregt_cursor = await self.__record_taregt_cursor.executemany(
            operation, seq_of_parameters
        )

        return self

    async def execute_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "AsyncRecordCursor[turu.core.cursor.GenericNewRowType, Parameters]":
        self.__record_taregt_cursor = await self.__record_taregt_cursor.execute_map(
            row_type, operation, parameters
        )

        return cast(AsyncRecordCursor, self)

    async def executemany_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "AsyncRecordCursor[turu.core.cursor.GenericNewRowType, Parameters]":
        self.__record_taregt_cursor = await self.__record_taregt_cursor.executemany_map(
            row_type, operation, seq_of_parameters
        )

        return cast(AsyncRecordCursor, self)

    async def fetchone(self) -> Optional[turu.core.cursor.GenericRowType]:
        row = await self.__record_taregt_cursor.fetchone()
        if row is not None:
            self._recorder.record([row])

        return row

    async def fetchmany(
        self, size: Optional[int] = None
    ) -> List[turu.core.cursor.GenericRowType]:
        rows = await self.__record_taregt_cursor.fetchmany(size)

        self._recorder.record(rows)

        return rows

    async def fetchall(self) -> List[turu.core.cursor.GenericRowType]:
        rows = await self.__record_taregt_cursor.fetchall()

        self._recorder.record(rows)

        return rows

    def __iter__(
        self,
    ) -> "AsyncRecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        return self

    async def __anext__(self) -> turu.core.cursor.GenericRowType:
        row = await self.__record_taregt_cursor.__anext__()

        self._recorder.record([row])

        return row

    def __getattr__(self, name):
        def _method_missing(*args):
            return args

        return getattr(self.__record_taregt_cursor, name, _method_missing)



from abc import abstractmethod
from typing import Optional


class TuruException(Exception):
    """Base class for all Turu exceptions."""


class TuruError(TuruException):
    """Raised when an error occurs in Turu."""

    @property
    @abstractmethod
    def message(self) -> str: ...

    def __str__(self) -> str:
        return self.message


class TuruRowTypeMismatchError(TuruError):
    """Raised when a row is not of the expected type."""

    def __init__(self, expected: Optional[type], actual: Optional[type]) -> None:
        self.expected = expected
        self.actual = actual

    @property
    def message(self) -> str:
        expected_type = self.expected.__name__ if self.expected else None
        actual_type = self.actual.__name__ if self.actual else None

        return f"Unsupported row type: Expected {expected_type}, got {actual_type}."


class TuruRowTypeNotSupportedError(TuruError):
    """Raised when a row type is not supported."""

    def __init__(self, row_type: Optional[type]) -> None:
        self.row_type = row_type

    @property
    def message(self) -> str:
        row_type = self.row_type.__name__ if self.row_type else None
        return f"Unsupported row type: {row_type}."


class TuruUnexpectedFetchError(TuruError):
    @property
    def message(self) -> str:
        return "Fetch is unexpected. use execute_*() to specify row type."



# NOTE: It's a draft. It's NOT ready for production use.

from typing import Protocol

from turu.core.protocols.async_cursor import AsyncCursorProtocol


class AsyncConnectionProtocol(Protocol):
    async def close(self) -> None:
        """Close the connection now."""
        ...

    async def commit(self) -> None:
        """Commit any pending transaction to the database."""
        ...

    async def rollback(self) -> None:
        """Roll back to the start of any pending transaction.

        Closing a connection without committing the changes first
        will cause an implicit rollback to be performed.
        """
        ...

    async def cursor(self) -> AsyncCursorProtocol:
        """Return a new Cursor Object using the connection."""
        ...



from .async_connection import MockAsyncConnection as MockAsyncConnection
from .async_cursor import MockAsyncCursor as MockAsyncCursor
from .connection import MockConnection as MockConnection
from .cursor import MockCursor as MockCursor
